import logging
from typing import Optional, List, Iterator, Sequence, Union, Tuple
import pandas as pd
from clickhouse_driver import Client
from gxkit_dbtools.parser.sql_parser import SQLParser
from gxkit_dbtools.client.base import BaseDBClient
from gxkit_dbtools.exception import DBConnectionError, SQLExecutionError


class ClickHouseClient(BaseDBClient):
    """
    ClickHouseClient - ClickHouse 原生客户端
    Version: 0.1.0
    """

    def __init__(self, host: str, port: int, user: str, password: str, database: str, **kwargs):
        self.client: Optional[Client] = None
        self.db_type: str = "clickhouse"
        self.connect(host, port, user, password, database, **kwargs)

    def connect(self, host: str, port: int, user: str, password: str, database: str, **kwargs) -> None:
        try:
            self.client = Client(host=host, port=port, user=user, password=password, database=database, **kwargs)
            self.client.execute("SELECT 1")
        except Exception as e:
            logging.error(f"Failed to connect to ClickHouse: {e}")
            raise DBConnectionError("dbtools.ClickHouseClient.connect", "clickhouse", str(e)) from e

    def execute(self, sql: str, stream: bool = False, batch_size: int = 10000, values: Optional[List[Tuple]] = None) -> \
            Union[pd.DataFrame, Iterator[pd.DataFrame], int, None]:
        if self.client is None:
            raise DBConnectionError("dbtools.ClickHouseClient.execute", "clickhouse", "Not connected")

        sql = sql.strip()

        if values is not None:
            if not sql.lower().endswith("values"):
                raise SQLExecutionError("dbtools.ClickHouseClient.execute", sql,
                                        "'values' mode requires SQL to end with 'VALUES'")

            fake_sql = f"{sql} (null)"
            try:
                fake_parser = SQLParser(fake_sql, db_type="clickhouse")
                sql_type = fake_parser.sql_type()
                if sql_type != "statement":
                    raise SQLExecutionError("dbtools.ClickHouseClient.execute", sql,
                                            f"'values' mode only supports statement SQL, but got sql_type={sql_type}")

            except Exception as e:
                raise SQLExecutionError("dbtools.ClickHouseClient.execute", sql,
                                        f"SQLParser failed on values-mode SQL: {e}") from e

            if stream:
                logging.warning("[dbtools.ClickHouseClient.execute] | 'stream' ignored when 'values' is provided.")

            return self._execute_core(sql, sql_type, values=values)

        parsed_sql = SQLParser(sql, db_type="clickhouse")
        sql_type = parsed_sql.sql_type()
        operation = parsed_sql.operation()

        if stream and sql_type == "statement":
            logging.warning("[dbtools.ClickHouseClient.execute] | Stream unsupported for statement SQL. Using default.")
            stream = False

        if stream and sql_type == "query":
            return self._stream_core(sql, operation, parsed_sql, batch_size)

        return self._execute_core(sql, sql_type)

    def executemany(self, sqls: Sequence[str], stream: bool = False, batch_size: int = 10000,
                    collect_results: bool = True, values: Optional[Union[List[Tuple], List[List[Tuple]]]] = None) -> \
            List[Union[pd.DataFrame, int, None]]:
        if isinstance(sqls, str):
            sqls = [sqls]
            if isinstance(values, list) and values and isinstance(values[0], tuple):
                values = [values]
            elif values is None:
                values = [None]
            elif not isinstance(values, list) or not isinstance(values[0], (tuple, list)):
                raise SQLExecutionError("ClickHouseClient.executemany", sqls[0], "'values' type invalid")

        if not sqls:
            raise SQLExecutionError("ClickHouseClient.executemany", str(sqls), "'sqls' must be a non-empty sequence")

        if values is not None and len(sqls) != len(values):
            raise SQLExecutionError("ClickHouseClient.executemany", str(sqls),
                                    "Length of 'sqls' and 'values' does not match")

        results = []
        try:
            for idx, sql in enumerate(sqls):
                parsed_sql = SQLParser(sql, db_type="clickhouse")
                sql_type = parsed_sql.sql_type()
                val = values[idx] if values else None
                result = self.execute(sql, stream=stream, batch_size=batch_size, values=val)
                if collect_results and sql_type == "query":
                    results.append(result)
            return results if collect_results else []
        except Exception as e:
            raise SQLExecutionError("dbtools.ClickHouseClient.executemany", "; ".join(sqls), str(e)) from e

    def close(self) -> None:
        if self.client:
            self.client.disconnect()
            self.client = None

    def _execute_core(self, sql: str, sql_type: str, values: Optional[List[Tuple]] = None) -> Union[
        pd.DataFrame, int, None]:
        try:
            if sql_type == "query":
                result, meta = self.client.execute(sql, with_column_types=True)
                if not result:
                    return None
                columns = [col[0] for col in meta]
                return pd.DataFrame(result, columns=columns)
            elif values is not None:
                self.client.execute(sql, values)
                return len(values)
            else:
                self.client.execute(sql)
                return 1
        except Exception as e:
            raise SQLExecutionError("dbtools.ClickHouseClient._execute_core", sql, str(e)) from e

    def _stream_core(self, sql: str, operation: str, parsed_sql: SQLParser, batch_size: int) -> Optional[
        Iterator[pd.DataFrame]]:
        try:
            columns = self._get_columns(operation, parsed_sql)
            if columns is None:
                return None
            iter_rows = self.client.execute_iter(sql)

            def generator():
                batch = []
                for row in iter_rows:
                    batch.append(row)
                    if len(batch) >= batch_size:
                        yield pd.DataFrame(batch, columns=columns)
                        batch.clear()
                if batch:
                    yield pd.DataFrame(batch, columns=columns)

            return generator()
        except Exception as e:
            raise SQLExecutionError("dbtools.ClickHouseClient._stream_core", sql, str(e)) from e

    def _get_columns(self, operation: str, parsed_sql: SQLParser) -> Optional[List[str]]:
        if operation == "select":
            column_sql = parsed_sql.change_segments({"limit": "10"})
            test, meta = self.client.execute(column_sql, with_column_types=True)
            return [col[0] for col in meta] if test else None
        else:
            return parsed_sql.columns(mode="alias")



from gxkit_dbtools.client.mysql_client import MySQLClient
from gxkit_dbtools.client.clickhouse_client import ClickHouseClient
from gxkit_dbtools.client.iotdb_client import IoTDBClient, IoTDBPoolClient

__all__ = [
    "MySQLClient",
    "ClickHouseClient",
    "IoTDBClient",
    "IoTDBPoolClient",
]



import logging
import time
import pandas as pd
import pymysql
from datetime import datetime, timedelta
from pymysql.cursors import Cursor, SSCursor
from typing import Optional, List, Iterator, Sequence, Union

from gxkit_dbtools.parser.sql_parser import SQLParser
from gxkit_dbtools.client.base import BaseDBClient
from gxkit_dbtools.exception import DBConnectionError, SQLExecutionError


class MySQLClient(BaseDBClient):
    """
    MySQLClient - MySQLClient 原生客户端
    Version: 0.1.0
    """

    def __init__(self, host: str, port: int, user: str, password: str, database: str, **kwargs):
        self.client: Optional[pymysql.connect] = None
        self._last_success_time: Optional[datetime] = None
        self._reconnect_delay: timedelta = timedelta(minutes=10)
        self.database = database
        self._connection_params = dict(host=host, port=port, user=user, password=password, database=database, **kwargs)
        self.connect()

    def connect(self) -> None:
        try:
            self.client = pymysql.connect(**self._connection_params)
            with self.client.cursor() as cursor:
                cursor.execute("SELECT 1")
            self._last_success_time = datetime.now()
        except Exception as e:
            raise DBConnectionError("dbtools.MySQLClient.connect", "mysql", str(e)) from e

    def execute(self, sql: str, stream: bool = False, batch_size: int = 10000, max_retry: int = 3) -> Union[
        pd.DataFrame, Iterator[pd.DataFrame], int, None]:
        if self.client is None:
            raise DBConnectionError("dbtools.MySQLClient.connect", "mysql", "Database not connected")
        parsed_sql = SQLParser(sql, db_type="mysql")
        sql_type = parsed_sql.sql_type()
        if stream and sql_type == "statement":
            logging.warning(
                "[dbtools.MySQLClient.execute] | Stream function unsupported for this SQL. Using normal execution.")
        elif stream and sql_type == "query":
            return self._stream_core_retry_wrapper(sql, batch_size, max_retry)
        return self._execute_core_retry_wrapper(sql, sql_type, max_retry, autocommit=True)

    def executemany(self, sqls: Sequence[str], wrapper: bool = True, collect_results: bool = True,
                    max_retry: int = 3) -> List[Union[pd.DataFrame, int, None]]:
        if not isinstance(sqls, Sequence) or isinstance(sqls, str) or not sqls:
            raise SQLExecutionError("dbtools.MySQLClient._execute_core", sqls, "unsupported sqls type")
        results = []
        try:
            if wrapper:
                self.client.begin()
            for sql in sqls:
                parsed_sql = SQLParser(sql, db_type="mysql")
                sql_type = parsed_sql.sql_type()
                result = self._execute_core_retry_wrapper(sql, sql_type, max_retry, autocommit=not wrapper)
                if collect_results:
                    results.append(result)
            if wrapper:
                self.client.commit()
            return results if collect_results else []
        except Exception as e:
            if wrapper:
                try:
                    self.client.rollback()
                except:
                    pass
            raise SQLExecutionError("dbtools.MySQLClient.executemany", ";".join(sqls), str(e)) from e

    def close(self):
        if self.client:
            try:
                self.client.close()
            except:
                pass
            self.client = None

    def _execute_core_retry_wrapper(self, sql: str, sql_type: str, max_retry: int, autocommit: bool) -> Union[
        pd.DataFrame, int, None]:
        for retry in range(max_retry):
            try:
                return self._execute_core(sql, sql_type, autocommit)
            except (pymysql.err.OperationalError, pymysql.err.InterfaceError) as e:
                logging.warning(f"[MySQLClient._execute_core] Retry {retry + 1}/{max_retry}: {e}")
                self.close()
                self.connect()
                time.sleep(0.5)
            except SQLExecutionError as e:
                raise SQLExecutionError("dbtools.MySQLClient._execute_core_retry_wrapper", sql, str(e)) from e
        raise SQLExecutionError("dbtools.MySQLClient._execute_core", sql, f"Failed after {max_retry} retries")

    def _execute_core(self, sql: str, sql_type: str, autocommit: bool) -> Union[pd.DataFrame, int, None]:
        self._check_and_reconnect()
        try:
            with self.client.cursor(Cursor) as cursor:
                if sql_type == "query":
                    cursor.execute(sql)
                    result = cursor.fetchall()
                    if not result:
                        return None
                    column_names = [desc[0] for desc in cursor.description]
                    self._last_success_time = datetime.now()
                    return pd.DataFrame(result, columns=column_names)
                else:
                    if autocommit:
                        self.client.begin()
                    cursor.execute(sql)
                    if autocommit:
                        self.client.commit()
                    self._last_success_time = datetime.now()
                    return 1
        except Exception as e:
            if sql_type == "statement" and autocommit:
                try:
                    self.client.rollback()
                except Exception as e:
                    logging.warning(f"Rollback failed: {e}")
            raise SQLExecutionError("dbtools.MySQLClient._execute_core", sql, str(e)) from e

    def _stream_core_retry_wrapper(self, sql: str, batch_size: int, max_retry: int) -> Optional[Iterator[pd.DataFrame]]:
        for retry in range(max_retry):
            try:
                return self._stream_core(sql, batch_size)
            except (pymysql.err.OperationalError, pymysql.err.InterfaceError) as e:
                logging.warning(f"[MySQLClient._stream_core] Retry {retry + 1}/{max_retry}: {e}")
                self.close()
                self.connect()
                time.sleep(0.5)
            except SQLExecutionError as e:
                raise SQLExecutionError("dbtools.MySQLClient._stream_core_retry_wrapper", sql, str(e)) from e
        raise SQLExecutionError("dbtools.MySQLClient._stream_core", sql, f"Failed after {max_retry} retries")

    def _stream_core(self, sql: str, batch_size: int) -> Optional[Iterator[pd.DataFrame]]:
        self._check_and_reconnect()
        cursor = None
        try:
            cursor = self.client.cursor(SSCursor)
            cursor.execute(sql)
            column_names = [desc[0] for desc in cursor.description]
            first_batch = cursor.fetchmany(batch_size)

            if not first_batch:
                cursor.close()
                return None
            self._last_success_time = datetime.now()

            def generator():
                try:
                    yield pd.DataFrame(first_batch, columns=column_names)
                    while True:
                        batch = cursor.fetchmany(batch_size)
                        if not batch:
                            break
                        self._last_success_time = datetime.now()
                        yield pd.DataFrame(batch, columns=column_names)
                finally:
                    cursor.close()

            return generator()
        except Exception as e:
            if cursor:
                cursor.close()
            raise SQLExecutionError("dbtools.MySQLClient._stream_core", sql, str(e)) from e

    def _check_and_reconnect(self):
        if self.client is None or self._last_success_time is None:
            return

        now = datetime.now()
        if now - self._last_success_time > self._reconnect_delay:
            try:
                logging.info("[dbtools.MySQLClient] Reconnecting due to inactivity timeout")
                self.client.ping(reconnect=True)
                self._last_success_time = now
            except Exception:
                logging.warning("[dbtools.MySQLClient] Ping failed, reconnecting from scratch...")
                self.close()
                self.connect()
                self._last_success_time = datetime.now()



"""
dbtools.parser 工具层
Version 0.1.0
"""

MULTIWORD_COMMAND_OPS = {
    # MySQL
    "show databases", "show tables", "show full tables", "show create table", "show table status", "show columns",
    "show columns from", "show full columns from", "show indexes", "show indexes from", "show keys from",
    "show processlist", "show status", "show engine status", "show engines", "show grants", "show master status",
    "show slave status", "show binary logs", "show binlog events", "show open tables", "show privileges",
    "show variables", "show global status", "show session status", "show global variables", "show session variables",
    "show warnings", "show errors",

    # ClickHouse
    "show tables", "show databases", "show create table", "show create database", "show settings", "show functions",
    "show clusters", "show macros", "show dictionaries", "show engines", "show indexes", "show policies", "show quota",
    "show roles", "show users", "show profile events", "show profile events histogram", "show profile events values",

    # IoTDB
    "show timeseries", "show child paths", "show child nodes", "show devices", "show storage group",
    "show storage groups", "show ttl", "show cluster", "show version", "show triggers", "show functions",
    "show schema templates", "show measurements", "show data type", "show region", "show regions", "show pipe",
    "show pipes", "show queries", "show count timeseries", "show count devices", "show paths using template",

    # 通用 command
    "explain analyze", "reset query cache", "flush privileges", "lock tables", "unlock tables", "set global",
    "set session", "set names", "set character set", "kill query", "kill connection", "analyze table", "repair table",
    "optimize table", "truncate table",
}

QUERY_OPS = {
    "select", "show", "explain", "desc", "describe", "use", "help", "analyze", "union", "intersect", "except", "with"
}

STATEMENT_OPS = {
    # 数据写入/更新
    "insert", "update", "delete", "replace", "merge",
    # DDL（数据定义语言）
    "create", "drop", "alter", "rename", "comment",
    "truncate", "truncatetable", "optimize", "check", "repair",
    # 权限管理
    "grant", "revoke",
    # 事务控制
    "begin", "start", "commit", "rollback", "savepoint", "release",
    # 系统设置
    "set", "reset", "flush", "kill",
    # 存储过程与执行计划
    "call", "prepare", "execute", "deallocate",
    # 并发控制
    "lock", "unlock",
    # 特殊数据库操作
    "vacuum", "load", "analyze table"
}

FIXED_COLUMNS = {
    "mysql": {
        "explain": [
            "id", "select_type", "table", "partitions", "type", "possible_keys",
            "key", "key_len", "ref", "rows", "filtered", "Extra"
        ],
        "show tables": ["Tables_in_{database}"],
        "show databases": ["Database"],
        "show columns": ["Field", "Type", "Null", "Key", "Default", "Extra"],
        "show columns from": ["Field", "Type", "Null", "Key", "Default", "Extra"],
        "show full columns from": ["Field", "Type", "Collation", "Null", "Key", "Default", "Extra", "Privileges",
                                   "Comment"],
        "show index": ["Table", "Non_unique", "Key_name", "Seq_in_index", "Column_name",
                       "Collation", "Cardinality", "Sub_part", "Packed", "Null", "Index_type",
                       "Comment", "Index_comment"],
        "show indexes": ["Table", "Non_unique", "Key_name", "Seq_in_index", "Column_name",
                         "Collation", "Cardinality", "Sub_part", "Packed", "Null", "Index_type",
                         "Comment", "Index_comment"],
        "show indexes from": ["Table", "Non_unique", "Key_name", "Seq_in_index", "Column_name",
                              "Collation", "Cardinality", "Sub_part", "Packed", "Null", "Index_type",
                              "Comment", "Index_comment"],
        "show keys from": ["Table", "Non_unique", "Key_name", "Seq_in_index", "Column_name",
                           "Collation", "Cardinality", "Sub_part", "Packed", "Null", "Index_type",
                           "Comment", "Index_comment"],
        "show processlist": ["Id", "User", "Host", "db", "Command", "Time", "State", "Info"],
        "show status": ["Variable_name", "Value"],
        "show global status": ["Variable_name", "Value"],
        "show session status": ["Variable_name", "Value"],
        "show variables": ["Variable_name", "Value"],
        "show global variables": ["Variable_name", "Value"],
        "show session variables": ["Variable_name", "Value"],
        "show privileges": ["Privilege", "Context", "Comment"],
        "show grants": ["Grants for user"],
        "show binary logs": ["Log_name", "File_size"],
        "show master status": ["File", "Position", "Binlog_Do_DB", "Binlog_Ignore_DB", "Executed_Gtid_Set"],
        "show slave status": ["Slave_IO_State", "Master_Host", "Master_User", "Master_Port", "Connect_Retry"],
        "show open tables": ["Database", "Table", "In_use", "Name_locked"],
        "show create table": ["Table", "Create Table"],
        "show engines": ["Engine", "Support", "Comment", "Transactions", "XA", "Savepoints"],
        "show engine status": ["Type", "Name", "Status"],
        "show errors": ["Level", "Code", "Message"],
        "show warnings": ["Level", "Code", "Message"],
        "show table status": ["Name", "Engine", "Version", "Row_format", "Rows", "Avg_row_length", "Data_length",
                              "Max_data_length", "Index_length", "Data_free", "Auto_increment", "Create_time",
                              "Update_time", "Check_time", "Collation", "Checksum", "Create_options", "Comment"],
        "describe": ["Field", "Type", "Null", "Key", "Default", "Extra"],
    },

    "clickhouse": {
        "explain": ["explain"],
        "show tables": ["name"],
        "show databases": ["name"],
        "show create table": ["statement"],
        "show create database": ["statement"],
        "show settings": ["name", "value", "changed", "description"],
        "show functions": ["name"],
        "show macros": ["macro", "value"],
        "show users": ["name"],
        "show roles": ["name"],
        "show quota": ["name"],
        "show policies": ["name"],
        "show clusters": ["cluster"],
        "show indexes": ["table", "name", "type", "expr", "granularity"],
        "show engines": ["name", "description"],
        "show dictionaries": ["name", "origin", "structure", "layout", "status"],
        "show profile events": ["event", "value"],
        "show profile events histogram": ["event", "bucket", "count"],
        "show profile events values": ["event", "value"],
        "describe": ["name", "type", "default_type", "default_expression", "comment", "codec_expression",
                     "ttl_expression"],
    },

    "iotdb": {
        "explain": ["operator", "plan", "details"],
        "show timeseries": ["timeseries", "alias", "database", "dataType", "encoding", "compression", "tags",
                            "attributes"],
        "show devices": ["devices", "isAligned"],
        "show child paths": ["childPaths"],
        "show child nodes": ["childNodes"],
        "show storage group": ["storageGroup"],
        "show storage groups": ["storageGroup"],
        "show ttl": ["storageGroup", "ttl"],
        "show cluster": ["NodeID", "Internal IP", "Port", "Role"],
        "show version": ["version"],
        "show triggers": ["triggerName", "status", "path", "className"],
        "show pipes": ["PipeName", "Role", "Status"],
        "show pipe": ["PipeName", "Role", "Status"],
        "show functions": ["function name", "class name"],
        "show schema templates": ["template name"],
        "show measurements": ["measurement"],
        "show data type": ["path", "dataType"],
        "show region": ["Region ID", "Type", "Status"],
        "show regions": ["Region ID", "Type", "Status"],
        "show queries": ["queryId", "statement", "startTime", "executionTime"],
        "show count timeseries": ["count"],
        "show count devices": ["count"],
        "show paths using template": ["path", "templateName"],
        "describe": ["timeseries", "alias", "database", "dataType", "encoding", "compression", "tags", "attributes"],
    }
}



from gxkit_dbtools.client import (
    MySQLClient,
    ClickHouseClient,
    IoTDBClient,
    IoTDBPoolClient,
)
from gxkit_dbtools.parser import SQLParser

__all__ = [
    "SQLParser",
    "MySQLClient",
    "ClickHouseClient",
    "IoTDBClient",
    "IoTDBPoolClient",
]



import re
import logging

import sqlglot
from sqlglot import expressions as exp
from sqlglot.errors import ParseError
from typing import List, Optional, Tuple, Dict, Union, Literal
from gxkit_dbtools.exception import SQLParseError
from gxkit_dbtools.parser.utils import MULTIWORD_COMMAND_OPS, QUERY_OPS, STATEMENT_OPS, FIXED_COLUMNS

logging.getLogger("sqlglot").setLevel(logging.ERROR)


class SQLParser:
    """
    SQLParser - SQL 超级解析器
    Version: 0.1.0
    """

    def __init__(self, sql: str, db_type: str = "mysql"):
        self.original_db_type = db_type
        self.db_type = db_type if db_type != "iotdb" else None
        self._sql_type: Optional[str] = None
        self._spec_flag: bool = False
        self._cache: Dict[str, Union[List[str], List[Tuple[str, str]], Dict[str, List[str]]]] = {}
        self._iotdb_paths: List[str] = []
        self.raw_sql, self.ast = self._parse(sql)
        self._operation, self._spec_flag = self._init_operation()

    def __str__(self) -> str:
        return self.raw_sql

    def __repr__(self) -> str:
        return f"<SQLParser: {self.operation().upper()} - {self.raw_sql}>"

    def _init_operation(self) -> Tuple[str, bool]:
        if self.ast.key and self.ast.key.lower() != "command":
            return self.ast.key.lower(), False
        return self._extract_spec_operation(), True

    def sql(self) -> str:
        sql = self.ast.sql(dialect=self.db_type)
        return self._restore_iotdb_paths(sql)

    def operation(self) -> str:
        return self._operation

    def sql_type(self) -> str:
        if self._sql_type is None:
            op_lower = self._operation
            if any(op_lower.startswith(q) for q in QUERY_OPS):
                self._sql_type = "query"
            elif any(op_lower.startswith(s) for s in STATEMENT_OPS):
                self._sql_type = "statement"
            else:
                self._sql_type = "unknown"
        return self._sql_type

    def tables(self, subquery: bool = False) -> List[str]:
        if self._spec_flag:
            if "tables:spec" not in self._cache:
                self._cache["tables:spec"] = self._extract_spec_table()
            return self._cache["tables:spec"]

        if "tables:main" not in self._cache or "tables:sub" not in self._cache:
            main_tables, subquery_tables = [], []
            for table_node in self.ast.find_all(exp.Table):
                catalog = table_node.args.get("catalog")
                db = table_node.args.get("db")
                name_expr = table_node.args.get("this")
                if not name_expr:
                    continue
                parts = []
                if catalog is not None:
                    parts.append(catalog.this)
                if db is not None:
                    parts.append(db.this)
                parts.append(name_expr.sql(dialect=self.db_type))
                table_name = ".".join(parts)
                table_name = table_name.replace("`", "").replace('"', "").replace("'", "")
                if table_node.find_ancestor(exp.Subquery):
                    subquery_tables.append(table_name)
                else:
                    main_tables.append(table_name)
            self._cache["tables:main"] = main_tables
            self._cache["tables:sub"] = subquery_tables
        return (
            self._cache["tables:main"] + self._cache["tables:sub"]
            if subquery
            else self._cache["tables:main"]
        )

    def columns(self, mode: Literal["raw", "alias", "full"] = "full", prefix: bool = False,
                db_type: Optional[str] = None) -> Union[List[str], List[Tuple[str, str]]]:
        if self.sql_type() != "query":
            return []

        db = (db_type or self.original_db_type or self.db_type or "").lower()
        key = f"columns:{mode}:prefix_{prefix}"

        if key not in self._cache:
            full_key = f"columns:full:prefix_{prefix}"
            if full_key not in self._cache:
                if self._operation == "select":
                    full_data = self._extract_select_columns(prefix)
                elif db in FIXED_COLUMNS and self._operation in FIXED_COLUMNS[db]:
                    full_data = [(col, col) for col in FIXED_COLUMNS[db][self._operation]]
                else:
                    full_data = []
                self._cache[full_key] = full_data
            else:
                full_data = self._cache[full_key]
            self._cache[f"columns:raw:prefix_{prefix}"] = [raw for raw, _ in full_data]
            self._cache[f"columns:alias:prefix_{prefix}"] = [alias for _, alias in full_data]

        return self._cache[key]

    def segments(self, seg_type: Union[str, List[str]] = "all", subquery: bool = False) -> Dict[str, List[str]]:
        def add_segment(segment: str, value: str):
            result.setdefault(segment, []).append(value)

        valid_seg_type = {
            "where": exp.Where,
            "group": exp.Group,
            "having": exp.Having,
            "order": exp.Order,
            "limit": exp.Limit,
            "offset": None
        }
        # 解析 seg_type 参数
        if seg_type == "all":
            seg_types = list(valid_seg_type.keys())
        elif isinstance(seg_type, str):
            seg_types = [seg_type.lower()]
        else:
            seg_types = [c.lower() for c in seg_type]
        seg_types = [c for c in seg_types if c in valid_seg_type]
        if not seg_types:
            raise SQLParseError("dbtools.SQLParser._parse", self.raw_sql,
                                f"No valid segment types provided. Supported: {list(valid_seg_type.keys())}")

        result: Dict[str, List[str]] = {}
        for seg in seg_types:
            cache_key = f"segments:{seg}:subquery_{subquery}"
            if cache_key in self._cache:
                for val in self._cache[cache_key]:
                    add_segment(seg, val)
                continue
            # offset 特判处理
            if seg == "offset":
                offsets = self._extract_offset(subquery=subquery)
                offsets = [offsets] if isinstance(offsets, str) else offsets or []
                self._cache[cache_key] = offsets
                for val in offsets:
                    add_segment(seg, val)
                continue
            # 生成并缓存其他子句
            segment_values: List[str] = []
            top_expr = self.ast.args.get(seg)
            if top_expr:
                segment_values.append(top_expr.sql(dialect=self.db_type))
                if seg == "limit" and isinstance(top_expr, exp.Limit):
                    offset_expr = top_expr.args.get("offset")
                    if offset_expr:
                        offset_clause = offset_expr.sql(dialect=self.db_type)
                        self._cache.setdefault(f"segments:offset:subquery_{subquery}", []).append(offset_clause)
            if subquery:
                top_id = id(top_expr) if top_expr else None
                for expr in self.ast.find_all(valid_seg_type[seg]):
                    if id(expr) == top_id:
                        continue
                    segment_values.append(expr.sql(dialect=self.db_type))
                    if seg == "limit" and isinstance(expr, exp.Limit):
                        offset_expr = expr.args.get("offset")
                        if offset_expr:
                            offset_clause = offset_expr.sql(dialect=self.db_type)
                            self._cache.setdefault(f"segments:offset:subquery_{subquery}", []).append(offset_clause)
            self._cache[cache_key] = segment_values
            for val in segment_values:
                add_segment(seg, val)

        return result

    def change_segments(self, replacements: Dict[str, str]) -> str:
        new_ast = self.ast.copy()

        # 每种操作对应的支持子句
        operation = self.operation().lower()
        allowed_segments_by_op = {
            "select": {"where", "group", "having", "order", "limit", "offset"},
            "insert": set(),
            "update": {"where"},
            "delete": {"where"},
        }
        allowed_segments = allowed_segments_by_op.get(operation, set())
        # 映射 segment -> AST 类型
        seg_expr_map = {
            "where": exp.Where,
            "group": exp.Group,
            "having": exp.Having,
            "order": exp.Order,
            "limit": exp.Limit,
        }
        for seg, raw_sql in replacements.items():
            seg = seg.lower()
            if seg not in allowed_segments:
                raise SQLParseError("change_segments", self.raw_sql,
                                    f"Segment '{seg}' is not allowed for operation '{operation}'")
            if seg == "offset":
                continue
            if seg not in seg_expr_map:
                raise SQLParseError("change_segments", self.raw_sql, f"Unsupported segment type: '{seg}'")

            # 去除 segment 前缀
            prefix = seg.upper().replace("GROUP", "GROUP BY").replace("ORDER", "ORDER BY")
            cleaned_sql = re.sub(rf"^\s*{seg}\b\s*", "", raw_sql.strip(), flags=re.IGNORECASE)
            inject_sql = f"SELECT * FROM dummy {prefix} {cleaned_sql}"
            try:
                inject_ast = sqlglot.parse_one(inject_sql)
            except Exception as e:
                raise SQLParseError("change_segments", inject_sql, f"Invalid SQL for segment '{seg}': {e}")

            seg_expr = inject_ast.args.get(seg)
            if not seg_expr:
                raise SQLParseError("change_segments", inject_sql, f"Segment '{seg}' parse failed from injected SQL.")

            new_ast.set(seg, seg_expr)
        # 初步生成 SQL
        new_sql = new_ast.sql(dialect=self.db_type)
        new_sql = self._restore_iotdb_paths(new_sql)
        # 单独处理 OFFSET
        if "offset" in replacements:
            if "offset" not in allowed_segments:
                raise SQLParseError("change_segments", self.raw_sql,
                                    f"Segment 'offset' is not allowed for operation '{operation}'")

            new_offset = re.sub(r"^\s*offset\b\s*", "", replacements["offset"].strip(), flags=re.IGNORECASE)
            if re.search(r"\bOFFSET\b\s+\d+", new_sql, flags=re.IGNORECASE):
                new_sql = re.sub(r"\bOFFSET\b\s+\d+", f"OFFSET {new_offset}", new_sql, flags=re.IGNORECASE)
            else:
                new_sql += f" OFFSET {new_offset}"
        return new_sql

    def change_columns(self, columns: Union[str, List[str]]) -> str:
        operation = self.operation().lower()
        if operation != "select":
            raise SQLParseError(
                "change_columns",
                self.raw_sql,
                "Only SELECT statements support changing columns",
            )

        cols_sql = ", ".join(columns) if isinstance(columns, list) else columns
        inject_sql = f"SELECT {cols_sql} FROM dummy"
        try:
            inject_ast = sqlglot.parse_one(inject_sql, dialect=self.db_type)
        except Exception as e:
            raise SQLParseError(
                "change_columns",
                inject_sql,
                f"Invalid column expressions: {e}",
            )

        exprs = inject_ast.args.get("expressions") or []
        if not exprs:
            raise SQLParseError(
                "change_columns",
                inject_sql,
                "Failed to parse column expressions",
            )

        new_ast = self.ast.copy()
        new_ast.set("expressions", exprs)
        new_sql = new_ast.sql(dialect=self.db_type)
        return self._restore_iotdb_paths(new_sql)

    def _parse(self, sql: str) -> tuple[str, exp.Expression]:
        try:
            sql_clean = self._clean_sql(sql)
            ast = sqlglot.parse_one(sql_clean, dialect=self.db_type)
        except ParseError as e:
            raise SQLParseError("dbtools.SQLParser._parse", sql, f"SQL syntax error: {e}") from e
        except Exception as e:
            raise SQLParseError("dbtools.SQLParser._parse", sql, f"SQL unknown error: {e}") from e
        return sql_clean, ast

    def _clean_sql(self, sql: str) -> str:
        # 鲁棒的去除注释
        sql = self._remove_sql_comments(sql)
        # 替换全角字符 → 半角
        sql = ''.join(
            chr(ord(char) - 0xFEE0) if 0xFF01 <= ord(char) <= 0xFF5E else char
            for char in sql
        )
        # 替换常见中文符号
        sql = sql.replace('，', ',').replace('）', ')').replace('（', '(').replace('：', ':') \
            .replace('‘', "'").replace('’', "'").replace('“', '"').replace('”', '"')
        # IoTDB 特殊路径预处理
        if self.original_db_type == "iotdb":
            sql, self._iotdb_paths = self._quote_iotdb_paths(sql)
        # 合并多余空格
        return re.sub(r"\s+", " ", sql).strip()

    def _extract_spec_operation(self) -> str:
        tokens = self.raw_sql.strip().lower().split()
        for n in range(min(4, len(tokens)), 0, -1):
            candidate = " ".join(tokens[:n])
            if candidate in MULTIWORD_COMMAND_OPS:
                return candidate
        return tokens[0] if tokens else "command"

    def _extract_spec_table(self) -> List[str]:
        tables = []
        # IoTDB 特殊表名匹配
        match = re.search(r"(root\.[\w\.\*]+)", self.raw_sql)
        if match:
            tables = [match.group(1)]
        # 普通匹配
        if not tables:
            operation = self._operation
            if operation.startswith("show"):
                tables = self.extract_after_keywords(self.raw_sql, [
                    "show create table", "show columns from", "show columns in", "show index from",
                    "show indexes from", "show keys from", "show full columns from", "show extended columns from"
                ])
            elif operation == "explain":
                match = re.search(r"from\s+([`\"']?\w+(?:[.`\"']\w+)?[`\"']?)", self.raw_sql, flags=re.IGNORECASE)
                if match:
                    tables = [match.group(1).replace("`", "").replace('"', "").replace("'", "")]
            elif operation in {"truncate", "truncatetable"}:
                tables = self.extract_after_keywords(self.raw_sql, ["truncate table", "truncate"])
            elif operation == "use":
                match = re.match(r"use\s+([`\"']?\w+[`\"']?)", self.raw_sql, flags=re.IGNORECASE)
                if match:
                    tables = [match.group(1).replace("`", "").replace('"', "").replace("'", "")]
        return tables

    def _extract_select_columns(self, prefix: bool) -> List[Tuple[str, str]]:
        result = []
        for expr in self.ast.expressions:
            if isinstance(expr, exp.Star):
                result.append(("*", "*"))
            elif isinstance(expr, exp.Alias):
                raw_expr = expr.this
                if isinstance(raw_expr, exp.Column):
                    raw_name = raw_expr.sql() if prefix else raw_expr.name
                else:
                    raw_name = raw_expr.sql()
                alias_name = expr.alias_or_name or raw_name
                result.append((raw_name, alias_name))
            elif isinstance(expr, exp.Column):
                raw_name = expr.sql() if prefix else expr.name
                alias_name = expr.name
                result.append((raw_name, alias_name))
            else:
                raw_name = expr.sql()
                alias_name = raw_name
                result.append((raw_name, alias_name))
        return result

    def _extract_offset(self, subquery: bool = False) -> Union[Optional[str], List[str]]:
        def get_offset_from_limit(expr: exp.Limit) -> Optional[str]:
            off = expr.args.get("offset")
            return off.sql(dialect=self.db_type) if off else None

        offsets: List[str] = []

        top_offset = self.ast.args.get("offset")
        if isinstance(top_offset, exp.Offset):
            offsets.append(top_offset.sql(dialect=self.db_type))

        top_limit = self.ast.args.get("limit")
        if isinstance(top_limit, exp.Limit):
            off_sql = get_offset_from_limit(top_limit)
            if off_sql:
                offsets.append(off_sql)

        if subquery:
            top_offset_id = id(top_offset) if isinstance(top_offset, exp.Offset) else None
            top_limit_id = id(top_limit) if isinstance(top_limit, exp.Limit) else None
            for expr in self.ast.find_all(exp.Offset):
                if id(expr) == top_offset_id:
                    continue
                offsets.append(expr.sql(dialect=self.db_type))
            for expr in self.ast.find_all(exp.Limit):
                if id(expr) == top_limit_id:
                    continue
                off_sql = get_offset_from_limit(expr)
                if off_sql:
                    offsets.append(off_sql)

        return offsets if subquery else (offsets[0] if offsets else None)

    @staticmethod
    def extract_after_keywords(sql: str, keywords: List[str]) -> List[str]:
        lowered_sql = sql.lower()
        for keyword in keywords:
            if keyword not in lowered_sql:
                continue
            pattern = rf"{keyword}\s+([`\"']?\w+(?:[.`\"']\w+)?[`\"']?)"
            match = re.search(pattern, sql, flags=re.IGNORECASE)
            if match:
                return [match.group(1).replace("`", "").replace('"', "").replace("'", "")]
        return []

    @staticmethod
    def _remove_sql_comments(sql: str) -> str:
        result = []
        i = 0
        in_single_quote = False
        in_double_quote = False
        in_line_comment = False
        in_block_comment = False
        length = len(sql)
        while i < length:
            char = sql[i]
            next_char = sql[i + 1] if i + 1 < length else ""
            # 进入字符串状态
            if not in_line_comment and not in_block_comment:
                if not in_double_quote and char == "'":
                    result.append(char)
                    in_single_quote = not in_single_quote
                    i += 1
                    continue
                elif not in_single_quote and char == '"':
                    result.append(char)
                    in_double_quote = not in_double_quote
                    i += 1
                    continue
            # 忽略字符串中的内容
            if in_single_quote or in_double_quote:
                result.append(char)
                i += 1
                continue
            # 开始 block 注释
            if not in_line_comment and not in_block_comment and char == "/" and next_char == "*":
                in_block_comment = True
                i += 2
                continue
            # 结束 block 注释
            if in_block_comment and char == "*" and next_char == "/":
                in_block_comment = False
                i += 2
                continue
            # 开始 line 注释
            if not in_line_comment and not in_block_comment and char == "-" and next_char == "-":
                in_line_comment = True
                i += 2
                continue
            # 结束 line 注释（遇到换行）
            if in_line_comment and char in "\n\r":
                in_line_comment = False
                result.append(char)
                i += 1
                continue
            # 默认追加字符（不在注释中）
            if not in_line_comment and not in_block_comment:
                result.append(char)
            i += 1
        return ''.join(result)

    @staticmethod
    def _quote_iotdb_paths(sql: str) -> Tuple[str, List[str]]:
        pattern = re.compile(
            r"(root(?:\.(?:[\w\*]+|`[^`]+`|\"[^\"]+\"|'[^']+'))+)"
        )
        result = []
        last_end = 0
        quoted: List[str] = []
        for match in pattern.finditer(sql):
            start, end = match.span()
            if start > 0 and sql[start - 1] in "'\"`":
                continue
            path = match.group(1)
            if (
                    "*" in path
                    or re.search(r"\.[0-9]", path)
                    or any(q in path for q in ("`", "'", '"'))
            ):
                result.append(sql[last_end:start])
                result.append(f"'{path}'")
                quoted.append(path)
                last_end = end
        result.append(sql[last_end:])
        return "".join(result), quoted

    def _restore_iotdb_paths(self, sql: str) -> str:
        if not self._iotdb_paths:
            return sql
        for path in self._iotdb_paths:
            sql = sql.replace(f'"{path}"', path)
            sql = sql.replace(f"'{path}'", path)
            sql = sql.replace(f'`{path}`', path)
        return sql



import logging
import re
import time
from datetime import datetime, timedelta
from typing import Optional, List, Tuple, Iterator, Sequence, Union

import pandas as pd
from iotdb import SessionPool
from iotdb.Session import Session
from iotdb.SessionPool import create_session_pool, PoolConfig

from gxkit_dbtools.client.base import BaseDBClient
from gxkit_dbtools.exception import DBConnectionError, SQLExecutionError
from gxkit_dbtools.parser.sql_parser import SQLParser

try:
    # 2.x path
    from iotdb.utils.exception import IoTDBConnectionException
except ImportError:
    # 1.x path
    try:
        from iotdb.utils import IoTDBConnectionException
    except ImportError:
        raise ImportError(
            "[IoTDBClient] Failed to import IoTDBConnectionException. Please ensure you have installed a compatible version of apache-iotdb"
        )


class BaseClient(BaseDBClient):
    """IoTDB 基础客户端"""

    def __init__(self, **connection_params):
        self.db_type = "iotdb"
        self._connection_params = connection_params

    # BaseDBClient abstract methods -------------------------------------------------
    def connect(self, *args, **kwargs):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def _get_session(self):
        raise NotImplementedError

    def _return_session(self, session):
        pass

    def execute(self, sql: str, auto_decision: bool = False, stream: bool = False, batch_size: Optional[int] = 10000,
                prefix_path: int = 1, max_retry: int = 3) -> Union[pd.DataFrame, Iterator[pd.DataFrame], int, None]:
        raise NotImplementedError

    def executemany(self, sqls: Sequence[str], auto_decision: bool = False, stream: bool = False,
                    batch_size: Optional[int] = 10000, collect_results: bool = True, prefix_path: int = 1,
                    max_retry: int = 3) -> List[Union[pd.DataFrame, int, None]]:
        if not isinstance(sqls, Sequence) or isinstance(sqls, str) or not sqls:
            raise SQLExecutionError("dbtools.IoTDBClient.executemany", str(sqls), "unsupported sqls type")

        results: List[Union[pd.DataFrame, int, None]] = []
        for sql in sqls:
            parsed_sql = SQLParser(sql, db_type="iotdb")
            sql_type = parsed_sql.sql_type()
            result = self.execute(
                sql,
                auto_decision=auto_decision,
                stream=stream,
                batch_size=batch_size,
                prefix_path=prefix_path,
                max_retry=max_retry,
            )
            if collect_results and sql_type == "query":
                results.append(result)

        return results if collect_results else []

    def _decision(self, session, sql: str, parsed_sql: SQLParser, prefix_path: int) -> Tuple[bool, int, int]:
        max_columns = 200
        max_rows = 100_000
        max_cells = 800_000
        max_limits = 20_000
        fallback_limit = 10_000
        small_limit_threshold = 2_000

        if parsed_sql.sql_type() != "query":
            return False, fallback_limit, 0
        if parsed_sql.operation() not in {"select", "union", "intersect", "except", "with"}:
            return False, fallback_limit, 0

        try:
            columns = self._get_columns(session, parsed_sql, prefix_path)
            column_count = len(columns)
            if column_count == 0:
                return False, fallback_limit, 0
            if column_count == 1 and re.match(r"(?i)\b(count|sum|avg|min|max)\s*\(.*\)", columns[0]):
                return False, 1, 1

            user_limit = parsed_sql.segments("limit").get("limit")
            if user_limit:
                try:
                    limit_value = int(user_limit[0].split()[1])
                    if limit_value <= small_limit_threshold:
                        return False, limit_value, 1
                except Exception:
                    pass

            count_column = columns[1].split(".")[-1] if len(columns) > 1 else columns[0].split(".")[-1]
            row_count = self._get_rows(session, parsed_sql, count_column)

            limit_value = min(max(1, max_cells // column_count), max_limits)
            if row_count:
                limit_value = min(limit_value, row_count)
            cell_score = (column_count * row_count) / max_cells
            need_page = cell_score > 1.0 or column_count > max_columns or row_count > max_rows
            page_count = (row_count // limit_value + (1 if row_count % limit_value else 0)) if row_count else 0

            return need_page, limit_value, page_count
        except Exception as e:
            raise SQLExecutionError("dbtools.IoTDBClient._decision", sql, str(e)) from e

    def _get_columns(self, session, parsed_sql: SQLParser, prefix_path: int) -> Optional[List[str]]:
        try:
            column_sql = parsed_sql.change_segments({"limit": "10"})
            df = self._query_dataframe(session, column_sql, prefix_path)
            return [] if df is None else list(df.columns)
        except Exception:
            return []

    def _get_rows(self, session, parsed_sql: SQLParser, column: str) -> int:
        limit = parsed_sql.segments("limit").get("limit")
        count_limit = None if not limit else int(limit[0].split()[1])
        try:
            count_sql = SQLParser(parsed_sql.change_columns(f"count({column})"), db_type="iotdb").sql()
        except Exception:
            inner_sql = parsed_sql.sql()
            count_sql = f"SELECT count(*) FROM ({inner_sql})"
        count_df = self._execute_core(session, count_sql, "query", 1)
        count_lines = int(count_df.iloc[0, 0]) if count_df is not None else 0
        return min(count_limit, count_lines) if count_limit else count_lines

    def _execute_stream_core(self, session, parsed_sql: SQLParser, batch_size: int, prefix_path: int) -> Optional[
        pd.DataFrame]:
        dfs = [df for df in self._paged_generator(session, parsed_sql, batch_size, prefix_path)]
        return pd.concat(dfs, ignore_index=True) if dfs else None

    def _stream_core(self, session, parsed_sql: SQLParser, batch_size: int, prefix_path: int) -> Optional[
        Iterator[pd.DataFrame]]:
        generator = self._paged_generator(session, parsed_sql, batch_size, prefix_path)
        try:
            first_batch = next(generator)
        except StopIteration:
            self._return_session(session)
            return None
        except Exception:
            self._return_session(session)
            raise

        def stream_generator():
            try:
                yield first_batch
                yield from generator
            finally:
                self._return_session(session)

        return stream_generator()

    def _paged_generator(self, session, parsed_sql: SQLParser, limit: int, prefix_path: int, col_time: str = "Time") -> \
            Iterator[pd.DataFrame]:
        seg_where = parsed_sql.segments("where").get("where")
        original_where = seg_where[0] if seg_where else None
        start_ts, end_ts = None, None

        if original_where:
            time_start_pattern = rf"(?i)\b{col_time}\s*(>=|>)\s*(\d+(?:\.\d+)?|'[^']+'|\"[^\"]+\")"
            time_end_pattern = rf"(?i)\b{col_time}\s*(<=|<)\s*(\d+(?:\.\d+)?|'[^']+'|\"[^\"]+\")"
            match_start = re.search(time_start_pattern, original_where)
            match_end = re.search(time_end_pattern, original_where)
            if match_start:
                start_ts = match_start.group(2).strip("'\"")
            if match_end:
                end_ts = match_end.group(2).strip("'\"")

        last_ts = start_ts
        prev_ts = None

        while True:
            logging.debug(f"[IoTDBClient._paged_generator] | last_ts: {last_ts}")
            final_where = None
            if original_where:
                cond = f"{col_time} > {last_ts}" if last_ts else None
                if cond:
                    time_gt_pattern = rf"(?i)\b{col_time}\s*(>=|>)\s*(\d+(?:\.\d+)?|'[^']+'|\"[^\"]+\")"
                    if re.search(time_gt_pattern, original_where):
                        final_where = re.sub(time_gt_pattern, cond, original_where, count=1)
                    else:
                        final_where = f"({original_where}) AND {cond}"
                else:
                    final_where = original_where
            elif last_ts:
                final_where = f"{col_time} > {last_ts}"

            replacements = {"limit": str(limit)}
            if final_where:
                replacements["where"] = final_where
            page_sql = parsed_sql.change_segments(replacements)
            logging.debug(f"[IoTDBClient._paged_generator] | final_where: {final_where}")
            logging.debug(f"[IoTDBClient._paged_generator] | page_sql: {page_sql}")

            df = self._execute_core(session, page_sql, "query", prefix_path)
            if df is None or df.empty:
                break

            df.columns = self._build_col_mapping(list(df.columns), prefix_path)
            yield df

            if len(df) < limit:
                break

            ts_max = df["timestamp"].max()
            if pd.notnull(ts_max):
                try:
                    last_ts = int(ts_max) + 1
                except Exception:
                    logging.warning(f"[IoTDBClient._paged_generator] | Invalid timestamp {ts_max}")
                    break
            else:
                break

            if last_ts == prev_ts:
                logging.warning("[IoTDBClient._paged_generator] | Timestamp not progressing. Breaking to avoid loop.")
                break

            if end_ts is not None and last_ts >= end_ts:
                break
            prev_ts = last_ts

    def _execute_core(self, session, sql: str, sql_type: str, prefix_path: int) -> Union[pd.DataFrame, int, None]:
        if sql_type != "query":
            session.execute_non_query_statement(sql)
            return 1
        return self._query_dataframe(session, sql, prefix_path)

    def _query_dataframe(self, session, sql: str, prefix_path: int) -> Optional[pd.DataFrame]:
        result = session.execute_query_statement(sql)
        df = result.todf()
        result.close_operation_handle()
        if df is None or df.empty:
            return None
        df.columns = self._build_col_mapping(list(df.columns), prefix_path)
        return df

    @staticmethod
    def _build_col_mapping(raw_cols: List[str], prefix_path: int) -> List[str]:
        def shorten(col: str) -> str:
            if col.lower() == "time":
                return "timestamp"

            if "(" in col and ")" in col:
                start = col.index("(")
                end = col.rindex(")")
                inner = col[start + 1: end]
                if "." in inner:
                    inner_short = ".".join(inner.split(".")[-prefix_path:]) if prefix_path > 0 else inner
                    return f"{col[: start + 1]}{inner_short}{col[end:]}"
                return col

            parts = col.split(".")
            return ".".join(parts[-prefix_path:]) if prefix_path > 0 else col

        return [shorten(col) for col in raw_cols]

    @staticmethod
    def _adjust_batch_size_for_limit(batch_size: int, limit_size: int, context: str = "execute") -> int:
        if batch_size is None or batch_size <= 0:
            return limit_size
        elif batch_size > limit_size:
            logging.warning(
                f"[IoTDBClient.{context}] | batch_size ({batch_size}) exceeds optimal limit ({limit_size}). Using limit_size instead."
            )
            return limit_size
        return batch_size


class IoTDBClient(BaseClient):
    """IoTDB 客户端(非线程池版)"""

    def __init__(self, host: str, port: int, user: str, password: str, **kwargs):
        super().__init__(host=host, port=port, user=user, password=password, **kwargs)
        self.session: Optional[Session] = None
        self._last_success_time: Optional[datetime] = None
        self._reconnect_delay: timedelta = timedelta(minutes=10)
        self.connect(host, port, user, password, **kwargs)

    def _check_and_reconnect(self):
        if self.session is None or self._last_success_time is None:
            return

        now = datetime.now()
        if now - self._last_success_time > self._reconnect_delay:
            try:
                logging.info("[dbtools.IoTDBClient] Reconnecting due to inactivity timeout")
                self.session.execute_query_statement("SHOW STORAGE GROUP").close_operation_handle()
                self._last_success_time = now
            except Exception:
                logging.warning("[dbtools.IoTDBClient] Ping failed, reconnecting from scratch...")
                try:
                    self.close()
                finally:
                    self.connect(**self._connection_params)
                self._last_success_time = datetime.now()

    def connect(self, host: str, port: int, user: str, password: str, **kwargs) -> None:
        try:
            self.session = Session(host, port, user, password, **kwargs)
            self.session.open(False)
            self.session.execute_query_statement("SHOW STORAGE GROUP").close_operation_handle()
            self._last_success_time = datetime.now()
        except Exception as e:
            raise DBConnectionError("dbtools.IoTDBClient.connect", "iotdb", str(e)) from e

    def _get_session(self):
        if self.session is None:
            raise DBConnectionError("dbtools.IoTDBClient.execute", "iotdb", "Database not connected")
        return self.session

    def execute(self, sql: str, auto_decision: bool = False, stream: bool = False, batch_size: Optional[int] = 10000,
                prefix_path: int = 1, max_retry: int = 3) -> Union[pd.DataFrame, Iterator[pd.DataFrame], int, None]:
        last_error = None
        for attempt in range(1, max_retry + 1):
            session = self._get_session()
            parsed_sql = SQLParser(sql, db_type="iotdb")
            sql_type = parsed_sql.sql_type()

            if sql_type == "statement" and stream:
                logging.warning(
                    "[dbtools.IoTDBClient.execute] | Stream function unsupported for this SQL. Using normal execution."
                )
                stream = False

            stream_started = False
            try:
                self._check_and_reconnect()
                if auto_decision and sql_type == "query":
                    need_page, limit_size, page_count = self._decision(session, sql, parsed_sql, prefix_path)
                    if page_count > 15:
                        logging.warning(
                            f"[IoTDBClient.execute] Query will be split into {page_count} pages. Adjust using IoTDBPoolClient."
                        )
                    if need_page:
                        if stream:
                            batch_size = self._adjust_batch_size_for_limit(batch_size, limit_size, context="execute")
                            result = self._stream_core(session, parsed_sql, batch_size, prefix_path)
                            stream_started = True
                            self._last_success_time = datetime.now()
                            return result
                        result = self._execute_stream_core(session, parsed_sql, limit_size, prefix_path)
                        self._last_success_time = datetime.now()
                        return result
                if stream and sql_type == "query":
                    result = self._stream_core(session, parsed_sql, batch_size, prefix_path)
                    stream_started = True
                    self._last_success_time = datetime.now()
                    return result
                result = self._execute_core(session, sql, sql_type, prefix_path)
                self._last_success_time = datetime.now()
                return result
            except IoTDBConnectionException as e:
                last_error = e
                logging.warning(
                    f"[IoTDBClient.execute] Retry {attempt}/{max_retry} after IoTDBConnectionException: {e}")
                self.close()
                self.connect(**self._connection_params)
                time.sleep(1)
            except Exception as e:
                if stream and sql_type == "query" and not stream_started:
                    self._return_session(session)
                raise SQLExecutionError("dbtools.IoTDBClient.execute", sql, str(e)) from e
            finally:
                if not (stream and sql_type == "query") or last_error is not None:
                    self._return_session(session)

        raise SQLExecutionError("dbtools.IoTDBClient.execute", sql, f"Failed after {max_retry} retries")

    def close(self) -> None:
        if self.session:
            try:
                self.session.close()
            except Exception:
                pass
            self.session = None


class IoTDBPoolClient(BaseClient):
    """IoTDB 客户端(线程池版)"""

    def __init__(self, host: str, port: int, user: str, password: str, max_pool_size: int = 10,
                 wait_timeout_in_ms: int = 3000, **kwargs):
        super().__init__(host=host, port=port, user=user, password=password, max_pool_size=max_pool_size,
                         wait_timeout_in_ms=wait_timeout_in_ms, **kwargs)
        self.pool: Optional[SessionPool] = None
        self.connect(host, port, user, password, max_pool_size, wait_timeout_in_ms, **kwargs)

    def connect(self, host: str, port: int, user: str, password: str, max_pool_size: int, wait_timeout_in_ms: int,
                retry: int = 3, **kwargs) -> None:
        try:
            config = PoolConfig(host=host, port=str(port), user_name=user, password=password, **kwargs)
            self.pool = create_session_pool(config, max_pool_size=max_pool_size, wait_timeout_in_ms=wait_timeout_in_ms)
            last_error = None

            for attempt in range(1, retry + 1):
                try:
                    session = self.pool.get_session()
                    try:
                        session.execute_query_statement("SHOW STORAGE GROUP").close_operation_handle()
                    finally:
                        self.pool.put_back(session)
                    return
                except Exception as e:
                    if type(e).__name__ == "IoTDBConnectionException":
                        last_error = e
                        if attempt < retry:
                            logging.warning(
                                f"[IoTDBClient.connect] Retry {attempt}/{retry} after IoTDBConnectionException: {e}"
                            )
                        else:
                            break
                    else:
                        raise DBConnectionError("dbtools.IoTDBClient.connect", "iotdb", str(e)) from e

            raise DBConnectionError("dbtools.IoTDBClient.connect", "iotdb", str(last_error)) from last_error

        except Exception as e:
            raise DBConnectionError("dbtools.IoTDBClient.connect", "iotdb", str(e)) from e

    def _get_session(self):
        if self.pool is None:
            raise DBConnectionError("dbtools.IoTDBClient.execute", "iotdb", "Database not connected")
        last_error = None
        for attempt in range(1, 4):
            try:
                return self.pool.get_session()
            except Exception as e:
                last_error = e
                logging.warning(
                    f"[IoTDBPoolClient._get_session] Retry {attempt}/3 after {type(e).__name__}: {e}"
                )
                time.sleep(0.5)
        raise DBConnectionError("dbtools.IoTDBClient.execute", "iotdb", str(last_error)) from last_error

    def _return_session(self, session):
        if self.pool is None:
            return
        last_error = None
        for attempt in range(1, 4):
            try:
                self.pool.put_back(session)
                return
            except Exception as e:
                last_error = e
                logging.warning(
                    f"[IoTDBPoolClient._return_session] Retry {attempt}/3 after {type(e).__name__}: {e}"
                )
                time.sleep(0.5)
        logging.error(
            f"[IoTDBPoolClient._return_session] Failed to return session after 3 retries: {last_error}"
        )

    def execute(self, sql: str, auto_decision: bool = False, stream: bool = False, batch_size: Optional[int] = 10000,
                prefix_path: int = 1, max_retry: int = 3) -> Union[pd.DataFrame, Iterator[pd.DataFrame], int, None]:
        last_error = None
        for attempt in range(1, max_retry + 1):
            session = self._get_session()
            parsed_sql = SQLParser(sql, db_type="iotdb")
            sql_type = parsed_sql.sql_type()

            if sql_type == "statement" and stream:
                logging.warning(
                    "[dbtools.IoTDBClient.execute] | Stream function unsupported for this SQL. Using normal execution."
                )
                stream = False

            stream_started = False
            try:
                if auto_decision and sql_type == "query":
                    need_page, limit_size, _ = self._decision(session, sql, parsed_sql, prefix_path)
                    if need_page:
                        if stream:
                            batch_size = self._adjust_batch_size_for_limit(batch_size, limit_size, context="execute")
                            result = self._stream_core(session, parsed_sql, batch_size, prefix_path)
                            stream_started = True
                            return result
                        return self._execute_stream_core(session, parsed_sql, limit_size, prefix_path)
                if stream and sql_type == "query":
                    result = self._stream_core(session, parsed_sql, batch_size, prefix_path)
                    stream_started = True
                    return result
                result = self._execute_core(session, sql, sql_type, prefix_path)
                return result
            except IoTDBConnectionException as e:
                last_error = e
                logging.warning(
                    f"[IoTDBClient.execute] Retry {attempt}/{max_retry} after IoTDBConnectionException: {e}")
                self.close()
                self.connect(**self._connection_params)
                time.sleep(1)
            except Exception as e:
                if stream and sql_type == "query" and not stream_started:
                    self._return_session(session)
                raise SQLExecutionError("dbtools.IoTDBClient.execute", sql, str(e)) from e
            finally:
                if not (stream and sql_type == "query") or last_error is not None:
                    self._return_session(session)

        raise SQLExecutionError("dbtools.IoTDBClient.execute", sql, f"Failed after {max_retry} retries")

    def close(self) -> None:
        if self.pool:
            try:
                self.pool.close()
            except Exception:
                pass
            self.pool = None
